@@ -9,8 +9,10 @@ import datawhys as dw
 from datawhys.core.frame import DataWhysFrame
 from datawhys.core.meta import BaseExploration
 from datawhys.core.series import DataWhysSeries
+from datawhys.datautils import apply_rule
 from datawhys.dd_transformer import DDTransformer
 from datawhys.error import NotEnoughPointsError
+from datawhys.metrics import partial_scores, score
 from datawhys.utils import meta as meta_utils, utilities
 
 
@@ -204,6 +206,52 @@ class Solver:
         else:
             return utilities.apply_rule_to_df(dataset, self.rule)
 
+    def get_node_data(self, dataset=None, node=None):
+        """
+        Recursively provides a dataset based on node in available Solver rule tree
+
+        Parameters
+        ----------
+
+        dataset: DataFrame or DataWhysFrame of shape (n_samples, n_features)
+            Where n_samples is the number of samples and n_features is the number of
+            features.
+
+        node: Specific node in rule_tree to retreive rule data
+            Node provided as List or Tuple available in Solver.rule_tree
+            after applying Solver.fit_tree().
+
+        Returns
+        -------
+        sample: pd.DataFrame
+            The sample of the original dataframe based on rule conditions
+
+        Examples
+        -------
+        >>> import datawhys as dw
+        >>> from datawhys.utils import graphviz
+
+        >>> solver = dw.Solver()
+        >>> # Exhaustively solve both in and out of rule subsets
+        >>> solver.fit_tree(X, y)
+
+        >>> # Get the filtered dataset for node (Level 3, IN rule)
+        >>> node = (0, "IN", 0, "IN", 1, "IN", 2, "IN", 3, "IN")
+        >>> # df assumed to be some dataset related to X, y
+        >>> solver.get_node_data(df, node)
+
+        """
+
+        def get_node(dataset, sequence, start=4):
+            seq = sequence[:start]
+            rule_data = apply_rule(dataset, self.rule_tree[seq]["rule"])
+            if start < len(sequence):
+                return get_node(rule_data, sequence, start + 2)
+            else:
+                return rule_data
+
+        return get_node(dataset, node)
+
     def fit(self, m_X, m_y, cv=None):
         """
         Fit the Solver with the provided data.
@@ -410,6 +458,149 @@ class Solver:
         evaluate(dwf)
         self.min_size_frac = min_size_frac_orig
 
+    def fit_tree(
+        self,
+        m_X,
+        m_y,
+        max_depth=None,
+        min_score=0,
+        balance=False,
+        balance_threshold=0.2,
+    ):
+        """
+        Fit the Solver with the provided data exhaustively solving both
+        in-rule and out-rule components.
+
+        Parameters
+        ----------
+
+        m_X: DataWhysFrame of shape (n_samples, n_features)
+            Where n_samples is the number of samples and n_features is the number of
+            features.
+
+        m_y: DataWhysSeries of shape (n_samples,)
+            Where n_samples is the number of samples and aligns with n_samples of m_X.
+
+        max_depth: Int, optional (default=None)
+            Sets a maximum number of levels in to split rules and incidentally the
+            subsamples defined by the rules.
+            Default `None` allows maximum number of splits.
+
+        min_score: Int, optional (default=0)
+            Sets a minimum Z-score found by rules as a threshold for rejecting rules.
+            If a rule is rejected, exploration down that branch will bet terminated.
+
+        balance: Bool, optional (default=False)
+            Allows the solver to create nested rules of roughly equal size in an
+            effort to balance the overall size of rules. In this process, the
+            Solver.min_size_frac is adjusted to allow subsequent rule size to
+            match the prior rule size as much as possible while maintaining rule
+            stability.
+
+        balance_threshold: float, optional (default=0.2)
+            If balance is set to True, the balance threshold allows setting how much
+            of a relative difference a subsequent rules size should be with respect to
+            the prior rule. This value can be as low as 0 to any multiple desired.
+
+            Example: balance_threshold = 0.2 implies that the subsequent rule size
+            should be within 20% (plus or minus) of the prior rule size
+
+        Examples
+        >>> import datawhys as dw
+        >>> from datawhys.utils import graphviz
+
+        >>> solver = dw.Solver()
+        >>> # Exhaustively solve both in and out of rule subsets
+        >>> solver.fit_tree(X, y)
+
+        >>> graphviz.visualize_rule_tree(solver.rule_tree)
+
+        """
+
+        explorable_vars = [col for col in m_X.columns if col != m_y.name]
+        m_X = m_X[explorable_vars].copy()
+
+        dwf = DataWhysFrame(pd.concat((DataWhysFrame(m_y), m_X), axis=1))
+
+        self.outcome = outcome = m_y.name
+        self.target = target = m_y.target_class
+
+        min_size_frac_orig = self.min_size_frac
+
+        self.rule_tree = dict()
+
+        def evaluate(data, depth=0, current="IN", prior=(0, "IN")):
+            def size_within_threshold(old_size, new_size, threshold=0.2):
+                if old_size == 0:
+                    return False
+                else:
+                    return np.abs(new_size / old_size - 1) <= threshold
+
+            if all(
+                (
+                    data.shape[0] > 1,
+                    data[outcome].unique().shape[0] > 1,
+                    depth <= max_depth if max_depth is not None else depth <= 1e6,
+                )
+            ):
+                min_size_orig = self.get_min_size_from_frac(data, outcome, target)
+                (
+                    self.rule,
+                    sample_stats,
+                    self.score,
+                    self.discarded_rules,
+                ) = self._run_solve(data, outcome, target)
+                self.score = score(data, outcome, target, self.rule)
+                self.size = utilities.apply_rule_to_df(data, self.rule).shape[0]
+                self.mean = sample_stats["mean"]
+
+                if self.score > min_score:
+                    key = prior + (depth, current)
+                    self.rule_tree[key] = {
+                        "rule": self.rule,
+                        "score": self.score,
+                        "size": self.size,
+                        "mean": self.mean,
+                        "partial": partial_scores(data, self.rule, outcome, target),
+                    }
+
+                    rule_data = self.get_rule_data(data).copy()
+                    xrule_data = self.get_rule_data(data, xrule=True).copy()
+
+                    try:
+                        if balance:
+                            if size_within_threshold(
+                                self.size, min_size_orig, balance_threshold
+                            ):
+                                self.set_min_size_frac_from_size(
+                                    rule_data, outcome, target, min_size_orig
+                                )
+                                evaluate(rule_data, depth + 1, "IN", key)
+                            else:
+                                pass
+                        else:
+                            evaluate(rule_data, depth + 1, "IN", key)
+                    except NotEnoughPointsError:
+                        pass
+                    try:
+                        if balance:
+                            if size_within_threshold(
+                                self.size, min_size_orig, balance_threshold
+                            ):
+                                self.set_min_size_frac_from_size(
+                                    xrule_data, outcome, target, min_size_orig
+                                )
+                                evaluate(xrule_data, depth + 1, "OUT", key)
+                            else:
+                                pass
+                        else:
+                            evaluate(xrule_data, depth + 1, "OUT", key)
+                    except NotEnoughPointsError:
+                        pass
+
+        evaluate(dwf)
+        self.min_size_frac = min_size_frac_orig
+
     def write_rule_tree(self, file_handle):
         """
         Export rule tree to json format.
